# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from __future__ import annotations

import datetime
import hashlib
import os
import pathlib
import pickle
import shutil
import tempfile
import uuid

from . import (
    gn,
    ty,
    walkproto,
)


class State:
    def __init__(self,
                 node: gn.Node,
                 workdir: ty.Union[pathlib.Path, str] = None,
                 ):
        self.node = node
        self.__workdir = pathlib.Path(workdir) if workdir else None
        self.__has_result = False
        self.__result = None

    def has_result(self) -> bool:
        return self.__has_result

    def get_result(self) -> ty.Any:
        if not self.__has_result:
            msg = f'state for node {self.node} has no valid result'
            raise RuntimeError(msg)
        return self.__result

    def set_result(self, result: ty.Any):
        self.__result = result
        self.__has_result = True

    def release(self):
        self.__result = None
        self.__has_result = False

    def is_up_to_date(self) -> bool:
        return False

    def get_timestamp(self) -> ty.Union[datetime.datetime, None]:
        return None

    def workdir(self) -> ty.Union[pathlib.Path, None]:
        if not self.__workdir:
            return None

        self.__workdir.mkdir(exist_ok=True, parents=True)
        return self.__workdir


class CachedState(State):
    def __init__(self,
                 node: gn.Node,
                 path: ty.Union[pathlib.Path, str] = None,
                 node_descriptor_path: ty.Union[pathlib.Path, str] = None,
                 workdir: ty.Union[pathlib.Path, str] = None,
                 check_saved_descriptor: bool = True
                 ):
        self.__path = pathlib.Path(path) if path else None
        self.__cached_is_up_to_date = None
        self.__cached_result_mtime = None
        self.__check_saved_descriptor = check_saved_descriptor

        self.__node_descriptor_path = None
        if node_descriptor_path:
            self.__node_descriptor_path = pathlib.Path(node_descriptor_path)

        if not workdir and self.__path:
            workdir = path / 'workdir'

        super().__init__(node, workdir)

    def __is_up_to_date(self) -> bool:
        # Do checks from the least to the most expensive

        # 1. Check if state directory exists.
        state_dir = self.__path / 'state'
        if not state_dir.is_dir():
            return False

        # 2. Check if any input path has it modification time greater than the
        #    state's timestamp.
        for p in self.node._pathins:
            pathin_mtime = datetime.datetime.fromtimestamp(
                pathlib.Path(p).stat().st_mtime,
                tz=datetime.timezone.utc,
            )
            if pathin_mtime > self.get_timestamp():
                return False

        # 3. Check if nodes this node depends on are up to date (the
        #    node_descriptor of such nodes are by definition smaller than this
        #    node's).
        for dep in self.node._get_dep_nodes():
            dep_state = get_state(dep)
            if not dep_state.is_up_to_date():
                return False
            if dep_state.get_timestamp() > self.get_timestamp():
                return False

        # 4. Finally, if this object was constructed with
        #    ``check_saved_descriptor=True``, compare this node's
        #    node_descriptor with the one saved in the state directory.
        if self.__check_saved_descriptor:
            node_descriptor = self.node._get_node_descriptor()
            node_descriptor_path = self.__node_descriptor_path
            if not node_descriptor_path:
                node_descriptor_path = state_dir / 'node_descriptor.pickle'
            with open(node_descriptor_path, 'rb') as f:
                saved_descriptor = pickle.load(f)
            if node_descriptor != saved_descriptor:
                return False

        return True

    def is_up_to_date(self) -> bool:
        if self.__cached_is_up_to_date is None:
            self.__cached_is_up_to_date = self.__is_up_to_date()
        return self.__cached_is_up_to_date

    def get_timestamp(self) -> datetime.datetime:
        if self.__cached_result_mtime is not None:
            return self.__cached_result_mtime
        result_path = self.__path / 'state' / 'result.pickle'
        result_mtime = datetime.datetime.fromtimestamp(
            result_path.stat().st_mtime,
            tz=datetime.timezone.utc,
        )
        self.__cached_result_mtime = result_mtime
        return result_mtime

    def has_result(self) -> bool:
        if super().has_result():
            return True
        result_path = self.__path / 'state' / 'result.pickle'
        return result_path.exists()

    def get_result(self) -> ty.Any:
        if not super().has_result():
            result_path = self.__path / 'state' / 'result.pickle'
            try:
                with open(result_path, 'rb') as f:
                    result = pickle.load(f)
            except FileNotFoundError:
                pass
            else:
                super().set_result(result)
        return super().get_result()

    def set_result(self, result: ty.Any):
        self.__path.mkdir(exist_ok=True, parents=True)

        tmpdir = pathlib.Path(tempfile.mkdtemp(dir=self.__path))
        try:
            if not self.__node_descriptor_path:
                with open(tmpdir / 'node_descriptor.pickle') as f:
                    node_descriptor = self.node._get_node_descriptor()
                    pickle.dump(node_descriptor, f)

            with open(tmpdir / 'result.pickle', 'wb') as f:
                pickle.dump(result, f)

            if (self.__path / 'state').exists():
                shutil.rmtree(self.__path / 'state')
            os.replace(tmpdir, self.__path / 'state')
        finally:
            # Remove temporary directory if it still exists
            if tmpdir.exists():
                shutil.rmtree(tmpdir)

        super().set_result(result)

    def release(self):
        self.__cached_is_up_to_date = None
        super().release()


class StateNamespace:
    def __init__(self, factory: Callable[[gn.Node], State] = State):
        self.__states = {}
        self.__node_descriptor_cache = {}
        self.factory = factory

    def get_state(self, node: gn.Node) -> State:
        current = self
        if node in self.__states:
            return self.__states[node]
        self.__states[node] = self.factory(node)
        return self.__states[node]

    def __enter__(self):
        global _current_namespace
        if _current_namespace:
            raise RuntimeError('there is already a state namespace in place')
        _current_namespace = self
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        global _current_namespace
        _current_namespace = None
        self.__cleanup()
        return False

    def __cleanup(self):
        for s in self.__states.values():
            s.release()
        self.__states = {}
        self.__node_descriptor_cache = {}

    def get_node_descriptor(self, node):
        return walkproto.node_descriptor(node, self.__node_descriptor_cache)


class CachedStateDB:
    def __init__(self, path: ty.Union[pathlib.Path, str] = 'db'):
        self.__path = pathlib.Path(path)

    def __call__(self, node: gn.Node) -> State:
        entry_dir = self.__find_entry_dir(node)
        return CachedState(
            node,
            path=entry_dir / 'statedir',
            node_descriptor_path=entry_dir / 'node_descriptor.pickle',
            # We use check_saved_descriptor=False here because we make sure
            # that the saved descriptor matches the node's descriptor in
            # __find_entry_dir()
            check_saved_descriptor=False,
        )

    def __find_entry_dir(self, node: gn.Node) -> pathlib.Path:
        if _current_namespace:
            node_descriptor = _current_namespace.get_node_descriptor(node)
        else:
            node_descriptor = node._get_node_descriptor()
        descriptor_bytes = pickle.dumps(node_descriptor)
        node_hash = hashlib.sha256(descriptor_bytes).hexdigest()

        bucket_dir = self.__path / 'entries' / node_hash
        for entry_dir in bucket_dir.glob('*'):
            with open(entry_dir / 'node_descriptor.pickle', 'rb') as f:
                entry_node_descriptor = pickle.load(f)
            if entry_node_descriptor == node_descriptor:
                return entry_dir

        # Create a new entry
        entry_id = str(uuid.uuid4())
        while (bucket_dir / entry_id).exists():
            entry_id = str(uuid.uuid4())
        entry_dir = bucket_dir / entry_id
        entry_dir.mkdir(exist_ok=True, parents=True)
        with open(entry_dir / 'node_descriptor.pickle', 'wb') as f:
            pickle.dump(node_descriptor, f)

        return entry_dir


def get_state(node: gn.Node) -> State:
    if not _current_namespace:
        raise RuntimeError('not in a state namespace context')
    return _current_namespace.get_state(node)


_current_namespace = None



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from __future__ import annotations

import collections
import pathlib

from . import ty


_op_types = []
def op_namedtuple(type_name, *names, **kw):
    t = collections.namedtuple(type_name, names, **kw)
    _op_types.append(t)
    return t


Data = op_namedtuple('Data', 'payload', 'id')


Value = op_namedtuple('Value', 'value')


GetItem = op_namedtuple('GetItem', 'obj', 'key')


GetAttr = op_namedtuple('GetAttr', 'obj', 'name')


Call = op_namedtuple('Call', 'fn', 'args', 'kwargs')


class PathIn(pathlib.PurePosixPath):
    pass


class PathOut(pathlib.PurePosixPath):
    pass


class _CTX:
    __slots__ = []

    singleton = None

    def __new__(cls):
        if _CTX.singleton:
            return _CTX.singleton
        r = super().__new__(cls)
        _CTX.singleton = r
        return r

CTX = _CTX()


class _UNSET:
    __slots__ = []

    singleton = None

    def __new__(cls):
        if _UNSET.singleton:
            return _UNSET.singleton
        r = super().__new__(cls)
        _UNSET.singleton = r
        return r


UNSET = _UNSET()


NodeOp = ty.Union[tuple(_op_types)]


def run_op(op):
    if isinstance(op, Data):
        return op.payload
    elif isinstance(op, Value):
        return op.value
    elif isinstance(op, GetItem):
        return op.obj[op.key]
    elif isinstance(op, GetAttr):
        return getattr(op.obj, op.name)
    elif isinstance(op, Call):
        return op.fn(*op.args, **op.kwargs)
    else:
        raise RuntimeError('unhandled operation, this is probably a bug')



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from __future__ import annotations

import pathlib
import logging
import pickle

from . import (
    mingraph,
    nodeop,
    nodestate,
    ty,
    walkproto,
)


logger = logging.getLogger()


NodeName = ty.Union[str, ty.Iterable[str]]


NodeRef = ty.Union[NodeName, 'Node']


class Node:
    def __init__(self,
            op: nodeop.NodeOp,
            name: str = None,
            always: bool = False,
            pathins: ty.Iterable[pathlib.Path] = tuple(),
            pathouts: ty.Iterable[pathlib.Path] = tuple(),
            graph: Graph = None,
            ):
        if not graph:
            if _graph_build_stack:
                graph = _graph_build_stack[-1]

        if name and '/' in name:
            raise ValueError(f'node name can not contain the slash character ("/"): {name}')

        self._op = op
        self._name = name
        self._always = always

        pathins = set(nodeop.PathIn(p) for p in pathins)
        pathouts = set(nodeop.PathOut(p) for p in pathouts)
        for evt in self.__op_walk():
            if evt.type == 'PathOut':
                pathouts.add(evt.value)
            elif evt.type == 'PathIn':
                pathins.add(evt.value)
        self._pathins = tuple(sorted(pathins))
        self._pathouts = tuple(sorted(pathouts))

        self.__graph = graph

        if self.__graph:
            self.__graph._Graph__add_node(self)
        else:
            # Validate things if node is outside of a graph
            if self._name:
                raise ValueError('named node is only allowed inside a graph')
            if self._pathouts:
                raise ValueError('pathouts are only allowed inside a graph')
            if self._pathins:
                raise ValueError('pathins are only allowed inside a graph')

    def _fullname(self):
        if not self._name:
            return None

        stack = [self._name]
        g = self.__graph
        root = self.__graph._Graph__root
        while g != root:
            stack.append(g.name)
            g = g._Graph__parent
        return '/'.join(reversed(stack))

    def _set(self, value):
        if not isinstance(self._op, nodeop.Value):
            raise ValueError('a value for a node can be set or unset only for Value operators')
        self._op = nodeop.Value(value)

    def _unset(self):
        self._set(nodeop.UNSET)

    def _result(self):
        return nodestate.get_state(self).get_result()

    def _must_run(self) -> bool:
        if self._always:
            return True
        return not nodestate.get_state(self).is_up_to_date()

    def __op_walk(self, op: nodeop.NodeOp = None) -> ty.Generator[walkproto.Event]:
        if not op:
            op = self._op
        yield from walkproto.walk(op)

    def _get_dep_nodes(self) -> ty.Generator[Node]:
        for evt in self.__op_walk():
            if evt.type == 'Node':
                yield evt.value
        for p in self._pathins:
            dep = self.__graph.path_producer(p)
            if dep:
                yield dep

    def _get_node_descriptor(self) -> tuple[walkproto.Event]:
        return walkproto.node_descriptor(self)

    def __getitem__(self, key):
        return Node(nodeop.GetItem(self, key))

    def __getattr__(self, name):
        if name in ('__getstate__', '__setstate__'):
            raise AttributeError('__setstate__ and __setstate__ are reserved for pickle')
        if isinstance(name, str) and name[0] == '_':
            msg = (
                f'failed to get attribute {name!r}: '
                'attributes starting with "_" are not supported via the dot operator (".")'
            )
            raise ValueError(msg)
        return Node(nodeop.GetAttr(self, name))

    def __call__(self, *args, **kwargs):
        return Node(nodeop.Call(self, args, kwargs))

    def __str__(self):
        return f'<{self._fullname()}>'

    def __repr__(self):
        return f'Node({self._op})'


class Graph:
    def __init__(self,
                name: str = None,
                parent: Graph = None,
             ):
        if name and '/' in name:
            raise ValueError(f'graph name can not contain the slash character ("/"): {name}')

        if not parent:
            if _graph_build_stack:
                parent = _graph_build_stack[-1]

        self.name = name
        self.__in_build_context = False
        self.__nodes = []
        self.__parent = parent
        self.__root = parent.__root if parent else self
        self.__graphs = []

        self.__name2node = {}
        """
        A dictionary mapping strings to either `Node` or `Graph` instances.
        """

        self.__pathout2node = {}
        """
        A dictionary mapping `nodeop.PathOut` instances to nodes that declare
        to produce them. This object should not be consulted directly, use the
        method `path_producer()` instead.
        """

        if self.__parent:
            self.__parent.__add_graph(self)

    def __enter__(self):
        if self.__in_build_context:
            raise RuntimeError('graph already in build context')
        _graph_build_stack.append(self)
        self.__in_build_context = True
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.__in_build_context = False
        _graph_build_stack.pop()

    def fullname(self) -> str:
        stack = []
        g = self
        while g != self.__root:
            stack.append(g.name)
            g = g.__parent
        return '/'.join(reversed(stack))

    def save(self, path: ty.Union[pathlib.Path, str]):
        if self.__in_build_context:
            msg = 'can not save a graph that is currently in build context'
            raise RuntimeError(msg)

        with open(path, 'wb') as f:
            CustomPickler(f).dump(self)

    @staticmethod
    def load(path: ty.Union[pathlib.Path, str]):
        with open(path, 'rb') as f:
            return pickle.load(f)

    def node(self, path: NodeName) -> ty.Union[Node, Graph]:
        if isinstance(path, str):
            parts = path.split('/')
        else:
            parts = tuple(path)

        if not parts:
            raise KeyError('received empty path as key')

        graph_names = parts[:-1]
        node_name = parts[-1]

        cur_graph = self
        for i, graph_name in enumerate(graph_names):
            if graph_name not in cur_graph.__name2node:
                partial_path = parts[:i]
                raise KeyError(f'graph at {partial_path!r} does not contain a child named {graph_name!r}')
            cur_graph = cur_graph.__name2node[graph_name]
            if not isinstance(cur_graph, Graph):
                partial_path = parts[:i+1]
                raise KeyError(f'element at {partial_path!r} is not a graph')

        if node_name not in cur_graph.__name2node:
            raise KeyError(f'graph at {graph_names!r} does not contain a node named {node_name!r}')

        return cur_graph.__name2node[node_name]

    def recurse_nodes(self) -> ty.Generator[Node]:
        yield from self.__nodes
        for g in self.__graphs:
            yield from g.recurse_nodes()

    def path_producer(self, path: pathlib.Path) -> Node:
        """
        Return the node that declares to produce the path `path` or None if there
        is no such node.
        """
        path = nodeop.PathOut(path)
        if path in self.__root.__pathout2node:
            return self.__root.__pathout2node[path]
        return _global_graph.__pathout2node.get(path)

    def mingraph(self,
                 unbounds: util.TargetsSpec,
                 targets: util.TargetsSpec,
                 ) -> Graph:
        return mingraph.mingraph(unbounds, targets, graph=self)

    def __add_graph(self, graph: Graph):
        if not graph.name:
            graph.name = f'graph-{len(self.__graphs)}'

        if graph.name in self.__name2node:
            msg = f'there is already a node named "{graph.name}"'
            raise ValueError(msg)
        self.__name2node[graph.name] = graph

        self.__graphs.append(graph)

    def __add_node(self, node: Node):
        if not node._name:
            if (isinstance(node._op, nodeop.Call)
                    and hasattr(node._op.fn, '__name__')):
                node._name = node._op.fn.__name__
                if node._name in self.__name2node:
                    node._name = f'{node._name}-{len(self.__nodes)}'
            else:
                node._name = f'unnamed-{len(self.__nodes)}'

        if node._name in self.__name2node:
            msg = f'there is already a node named "{node._name}"'
            raise ValueError(msg)
        self.__name2node[node._name] = node

        for p in node._pathouts:
            if self.path_producer(p):
                msg = f'found multiple nodes declaring to produce {p}'
                raise ValueError(msg)
            self.__root.__pathout2node[p] = node

        self.__nodes.append(node)


class CustomPickler(pickle.Pickler):
    def reducer_override(self, obj):
        if getattr(obj, '__module__', '') == '__main__':
            msg = (
                f'the object {obj} is defined in the __main__ module, '
                'you may run into issues if loading this graph from another module'
            )
            logger.warning(msg)
        return NotImplemented

_graph_build_stack = []
_global_graph = Graph()



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from __future__ import annotations

import collections

from . import (
    gn,
    ty,
)


def topological_sort(target_nodes: ty.Iterable[gn.Node],
                     ) -> tuple[list[gn.Node], collections.Counter]:
    visited = set()
    visiting = set()
    sorted_nodes = []
    path = []
    stack = [(node, None) for node in target_nodes]
    dependant_counts = collections.Counter()
    while stack:
        node, state = stack.pop()

        if state is None:
            if node in visited:
                # This node and dependencies have already been added to the
                # execution list
                continue

            path.append(node)

            if node in visiting:
                # If I am still visiting this node, this means a circular
                # dependency has been found
                path_str = ' <- '.join(str(n) for n in reversed(path))
                raise Exception(f'circular dependency found between nodes: {path_str}')

            # Get dependencies, push back to stack and mark node as
            # visiting
            deps = list(set(node._get_dep_nodes()))

            for dep in deps:
                dependant_counts[dep] += 1

            stack.append((node, deps))
            visiting.add(node)
        else:
            deps = state
            # If there are no more dependencies to be added to the
            # execution list, the node is ready to be added. If not, then
            # push the next node to the stack.
            if not deps:
                sorted_nodes.append(node)
                visited.add(node)
                visiting.remove(node)
                path.pop()
            else:
                dep = deps.pop()
                stack.append((node, deps))
                stack.append((dep, None))

    return sorted_nodes, dependant_counts


TargetsSpec = ty.Union[
    'gn.NodeRef',
    ty.Callable,
    ty.Sequence['gn.NodeRef'],
    ty.Mapping[str, 'gn.NodeRef'],
]


ParsedTargetsSpec = ty.Union[
    'gn.Node',
    tuple['gn.Node'],
    dict[str, 'gn.NodeRef'],
]


def parse_targets(targets: TargetsSpec,
                  graph: gn.Graph,
                  ) -> tuple[set[fn.Node], ParsedTargetsSpec]:
    if targets is None:
        if not graph:
            raise ValueError('graph is required when targets is None')
        targets = tuple(graph.recurse_nodes())

    def get_node(ref: gn.NodeRef, graph: ty.Union[gn.Graph, None]) -> gn.Node:
        if isinstance(ref, gn.Node):
            return ref
        else:
            if not graph:
                raise ValueError('graph is required when node reference is not a Node instance')

            node = graph.node(ref)

            if not isinstance(node, gn.Node):
                raise RuntimeError('node for ref {ref!r} not found')

            return node

    if isinstance(targets, ty.Mapping):
        targets = {
            k: get_node(v, graph) for k, v in targets.items()
        }
        nodes = set(targets.values())
    elif isinstance(targets, ty.Sequence) and not isinstance(targets, str):
        targets = tuple(get_node(t, graph) for t in targets)
        nodes = set(targets)
    elif callable(targets) and not isinstance(targets, gn.Node):
        if not graph:
            raise ValueError('graph is required when targets is a callable')
        targets = tuple(n for n in graph.recurse_nodes() if targets(n))
        nodes = set(targets)
    else:
        # The remaining possible type to expect is a NodeRef.
        # NOTE: the variable targets here refers to a single node despite its
        # name.
        targets = get_node(targets, graph)
        nodes = {targets}
    return nodes, targets



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""
Utility module to group commonly used types for type annotation.
"""
from collections.abc import (
    Callable,
    Generator,
    Iterable,
    Sequence,
    Mapping,
)
from typing import (
    Any,
    Union,
)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""
This module defines event types and functions used by `Node._op_walk()` and
`Node._get_node_descriptor()`.
"""
from __future__ import annotations

import collections
import inspect
import pathlib

from . import (
    gn,
    grun,
    nodeop,
    ty,
)


_event_types = []

def event_type(name: str, *fields: tuple[str]) -> type[collections.namedtuple]:
    """
    A factory function for creating types.

    The resulting type is a namedtuple with "type" as the first field name and
    `fields` as the remaining field names. When creating a new instance of the
    named tuple, the value of the first field will automatically be set to
    `name`.

    The rationale behind such a behavior is that, since
    `Node._get_node_descriptor()` is intended to be used as a way of uniquely
    identifying the node, it is important emit the type of the event together
    with its arguments.
    """
    fields = ('type', *fields)
    base = collections.namedtuple(name, fields)

    def __new__(cls, *k, **kw):
        k = (name, *k)
        return base.__new__(cls, *k, **kw)

    def __getnewargs__(self):
        return self[1:]

    namespace = {
        '__new__': __new__,
        '__getnewargs__': __getnewargs__,
    }

    t = type(name, (base,), namespace)
    _event_types.append(t)
    return t


# The types below are used by `Node._op_walk()` and
# `Node._get_node_descriptor()`
ValueId = event_type('ValueId', 'id')
Ref = event_type('Ref', 'id')
OpType = event_type('OpType', 'value')
DataOp = event_type('DataOp', 'value')
PathOut = event_type('PathOut', 'value')
PathIn = event_type('PathIn', 'value')
Node = event_type('Node', 'value')
CTX = event_type('CTX')
UNSET = event_type('UNSET')
Tuple = event_type('Tuple', 'size')
List = event_type('List', 'size')
Dict = event_type('Dict', 'keys')
Other = event_type('Other', 'value')


# The types below are used by `Node._get_node_descriptor()`
OpTypeDescriptor = event_type('OpTypeDescriptor', 'name')
CallableDescriptor = event_type('CallableDescriptor', 'source')
PathinsDescriptor = event_type('PathinsDescriptor', 'paths')
PathoutsDescriptor = event_type('PathoutsDescriptor', 'paths')


# Define Event as the Union of all event types
Event = ty.Union[tuple(_event_types)]


def walk(op: nodeop.NodeOp):
    refs = {}
    yield OpType(type(op))
    if isinstance(op, nodeop.Data):
        yield DataOp(op)
    else:
        for v in op:
            yield from walk_value(v, refs)


def walk_value(value: ty.Any, refs: dict) -> ty.Generator[Event]:
    if id(value) in refs:
        yield Ref(refs[id(value)])
        return

    refs[id(value)] = len(refs)
    yield ValueId(refs[id(value)])

    if isinstance(value, nodeop.PathOut):
        yield PathOut(value)
    elif isinstance(value, nodeop.PathIn):
        yield PathIn(value)
    elif isinstance(value, gn.Node):
        yield Node(value)
    elif value is nodeop.CTX:
        yield CTX()
    elif value is nodeop.UNSET:
        yield UNSET()
    elif type(value) == list:
        yield List(size=len(value))
        for v in value:
            yield from walk_value(v, refs)
    elif type(value) == tuple:
        yield Tuple(size=len(value))
        for v in value:
            yield from walk_value(v, refs)
    elif type(value) == dict:
        keys = tuple(value)
        yield Dict(keys=keys)
        for k in keys:
            yield from walk_value(value[k], refs)
    else:
        yield Other(value)


def node_descriptor(node: gn.Node,
                    cache: dict[gn.Node, tuple[Event]] = None,
                    ) -> tuple[Event]:
    if (cache is not None
            and not isinstance(node._op, nodeop.Value)
            and node in cache):
        return cache[node]

    op = node._op

    if isinstance(op, nodeop.Data):
        # For Data operations, the attribute "id" of the Data operation
        # identifies the data if present. In that case, we remove the
        # payload to make things light and fast.
        if op.id:
            op = op._replace(payload=None)

    desc = []

    desc.append(PathinsDescriptor(node._pathins))
    desc.append(PathoutsDescriptor(node._pathouts))
    for evt in walk(op):
        if isinstance(evt, Node):
            n = evt.value
            evt = evt._replace(value=None)
            desc.append(evt)
            desc.append(node_descriptor(n, cache))
        elif (isinstance(evt, Other)
              and callable(evt.value)
              and not inspect.isbuiltin(evt.value)):
            fn = evt.value
            desc.append(CallableDescriptor(
                # NOTE: It would be nice if we could add information from the
                # function's closure and default arguments as well. An issue
                # with that is that it will be common for some unpickable
                # objects to appear.
                source=inspect.getsource(fn),
            ))
        else:
            desc.append(evt)
    desc = tuple(desc)
    if cache is not None and not isinstance(node._op, nodeop.Value):
        cache[node] = desc
    return desc


def resolve_op(op: nodeop.NodeOp,
               ctx: grun.NodeContext,
               custom_atom_resolver: ty.Callable = None,
               ) -> nodeop.NodeOp:
    return OpResolver(op, ctx, custom_atom_resolver).resolve()


class _UNRESOLVED:
    __slots__ = []

    singleton = None

    def __new__(cls):
        if _UNRESOLVED.singleton:
            return _UNRESOLVED.singleton
        r = super().__new__(cls)
        _UNRESOLVED.singleton = r
        return r


UNRESOLVED = _UNRESOLVED()
"""
Special value to be returned by custom atom resolvers
(`OpResolver.custom_atom_resolver`) when default behavior is expected.
"""


class OpResolver:
    def __init__(self,
                 op: nodeop.NodeOp,
                 ctx: grun.NodeContext,
                 custom_atom_resolver: ty.Callable = None,
                 ):
        self.__ctx = ctx
        self.__op = op
        self.__cache = {}
        self.custom_atom_resolver = custom_atom_resolver

    def resolve(self):
        if isinstance(self.__op, nodeop.Data):
            return self.__op
        self.__events = walk(self.__op)
        return self.__resolve_op()

    def __resolve_op(self) -> nodeop.NodeOp:
        op_type = next(self.__events).value
        num_args = len(op_type._fields)
        args = [None] * num_args
        for i in range(num_args):
            args[i] = self.__resolve_value()
        return op_type(*args)

    def __resolve_value(self) -> ty.Any:
        evt = next(self.__events)
        if isinstance(evt, Ref):
            return self.__cache[evt.id]

        # Otherwise, evt will be a ValueId
        value_id = evt.id

        # Now get the next event, which describes the value
        evt = next(self.__events)
        if isinstance(evt, (List, Tuple)):
            resolved = [None] * evt.size
            for i in range(evt.size):
                resolved[i] = self.__resolve_value()
            if isinstance(evt, Tuple):
                resolved = tuple(resolved)
        elif isinstance(evt, Dict):
            keys = evt.keys
            resolved = {}
            for k in keys:
                resolved[k] = self.__resolve_value()
        else:
            resolved = self.__resolve_atom(evt)

        self.__cache[value_id] = resolved
        return resolved

    def __resolve_atom(self, evt):
        if self.custom_atom_resolver:
            resolved = self.custom_atom_resolver(evt)
            if resolved is not UNRESOLVED:
                return resolved

        if isinstance(evt, (PathOut, PathIn)):
            resolved = pathlib.Path(evt.value)
        elif isinstance(evt, Node):
            resolved = evt.value._result()
        elif isinstance(evt, CTX):
            resolved = self.__ctx
        elif isinstance(evt, UNSET):
            resolved = None
        elif isinstance(evt, Other):
            resolved = evt.value
        else:
            raise RuntimeError(f'unhandled value event, this is probably a bug: {evt!r}')
        return resolved



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""
"""
from __future__ import annotations

import contextlib

from . import (
    gn,
    nodeop,
    nodestate,
    ty,
    util,
    walkproto,
)


RunResult = ty.Union[
    'ty.Any',
    tuple,
    dict[str, 'ty.Any'],
]


class Runner:
    def run(self,
            targets: util.TargetsSpec = None,
            graph: gn.Graph = None,
            ns: nodestate.StateNamespace = None,
            ) -> RunResult:
        target_nodes, targets = util.parse_targets(targets, graph)

        # Get nodes to be executed
        nodes_to_run, dependant_counts = util.topological_sort(target_nodes)

        # Get or create the state namespace
        node_state_ctx = contextlib.nullcontext()
        if not nodestate._current_namespace:
            if not ns:
                ns = nodestate.StateNamespace()
            node_state_ctx = ns

        # Run nodes
        with node_state_ctx:
            for node in nodes_to_run:
                if not node._must_run():
                    continue
                ctx = NodeContext(node)
                resolved_op = walkproto.resolve_op(node._op, ctx)
                result = nodeop.run_op(resolved_op)
                nodestate.get_state(node).set_result(result)
                for dep in node._get_dep_nodes():
                    dependant_counts[dep] -= 1
                    if not dependant_counts[dep] and dep not in target_nodes:
                        nodestate.get_state(dep).release()

            # Generate return value
            if isinstance(targets, dict):
                return_value = {
                    k: n._result() for k, n in targets.items()
                }
            elif isinstance(targets, tuple):
                return_value = tuple(n._result() for n in targets)
            else:
                return_value = targets._result()

        return return_value


class NodeContext:
    def __init__(self, node: gn.Node):
        self.__node = node

    def workdir(self):
        return nodestate.get_state(self.__node).workdir()



import pathlib

import setuptools

setuptools.setup(
    name='yape',
    version='0.1.0',
    url='https://github.com/guludo/yape',
    description='Yape - Yet another pipeline executor',
    long_description=(pathlib.Path(__file__).parent / 'README.rst').read_text(),
    long_description_content_type='text/x-rst',
    packages=['yape'],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)',
        'Programming Language :: Python :: 3',
    ],
    keywords='pipeline',
)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from __future__ import annotations

from . import (
    gn,
    nodeop,
    walkproto,
    ty,
    util,
)


def mingraph(unbounds: util.TargetsSpec,
             targets: util.TargetsSpec,
             graph: gn.Graph = None,
             dest: gn.Graph = None,
             ) -> gn.Graph:
    return MingraphBuilder(unbounds, targets, graph, dest).build()


class MingraphBuilder:
    def __init__(self,
                 unbounds: util.TargetsSpec,
                 targets: util.TargetsSpec,
                 graph: fn.Graph = None,
                 dest: fn.Graph = None,
                 ):
        to_be_unbound, unbounds = util.parse_targets(unbounds, graph)
        target_nodes, targets = util.parse_targets(targets, graph)

        self.__to_be_unbound = to_be_unbound
        self.__unbounds = unbounds
        self.__target_nodes = target_nodes
        self.__targets = targets
        self.__dest = dest or gn.Graph()
        self.__new_nodes_cache = {}

        # Dictionary that will tell whether a node transitively depends on a
        # node to be unbound. This is used as cache for the method
        # __depends_on_unbound()
        self.__depends_on_unbound_cache = {}

        self.__generate_future_names()

    def build(self):
        with self.__dest:
            for node in self.__target_nodes:
                self.__get_new_node(node)
        return self.__dest

    def __get_new_node(self, node):
        if node in self.__new_nodes_cache:
            return self.__new_nodes_cache[node]

        if self.__depends_on_unbound(node) and node in self.__to_be_unbound:
            raise RuntimeError(
                f'node {node} is unbound and depends on an unbound'
            )
        elif self.__depends_on_unbound(node):
            op = walkproto.resolve_op(
                node._op,
                None,
                self.__custom_atom_resolver,
            )
        elif node in self.__to_be_unbound:
            op = nodeop.Value(nodeop.UNSET)
        else:
            op = nodeop.Data(node._result(), id=None)

        new_node = gn.Node(op, name=self.__future_names.get(node))
        self.__new_nodes_cache[node] = new_node
        return new_node

    def __depends_on_unbound(self, node: ng.Node, visited=None) -> bool:
        if node in self.__depends_on_unbound_cache:
            return self.__depends_on_unbound_cache[node]

        if visited is None:
            visited = set()

        if node in visited:
            raise RuntimeError('cycle detected')
        visited.add(node)

        r = False
        for dep in node._get_dep_nodes():
            if dep in self.__to_be_unbound:
                r = True
                break

            if self.__depends_on_unbound(dep, visited):
                r =  True
                break

        self.__depends_on_unbound_cache[node] = r
        return r

    def __custom_atom_resolver(self, evt: walkproto.Event) -> ty.Any:
        if isinstance(evt, walkproto.Node):
            return self.__get_new_node(evt.value)
        elif isinstance(evt, (walkproto.PathOut, walkproto.PathIn)):
            return evt.value
        elif isinstance(evt, walkproto.CTX):
            return nodeop.CTX
        else:
            return walkproto.UNRESOLVED

    def __generate_future_names(self):
        self.__future_names = {}

        if isinstance(self.__unbounds, dict):
            for name, node in self.__unbounds.items():
                if node in self.__future_names:
                    raise ValueError(
                        f'multiple future names for node {node}: '
                        f'{name!r} and {self.__future_names[node]!r}'
                    )
                self.__future_names[node] = name

        if isinstance(self.__targets, dict):
            for name, node in self.__targets.items():
                if isinstance(self.__unbounds, dict) \
                        and name in self.__unbounds:
                    raise ValueError(
                        f'the target node {node} has a future name ({name!r}) '
                        f'already used for an unbound node'
                    )
                if node in self.__future_names:
                    raise ValueError(
                        f'multiple future names for node {node}: '
                        f'{name!r} and {self.__future_names[node]!r}'
                    )
                self.__future_names[node] = name



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from __future__ import annotations

from . import (
    gn,
    grun,
    mingraph,
    nodeop,
    nodestate,
    ty,
)


PathIn = nodeop.PathIn


PathOut = nodeop.PathOut


CTX = nodeop.CTX


UNSET = nodeop.UNSET


State = nodestate.State


CachedState = nodestate.CachedState


StateNamespace = nodestate.StateNamespace


CachedStateDB = nodestate.CachedStateDB



def graph(**kw) -> gn.Graph:
    return gn.Graph(**kw)


mingraph = mingraph.mingraph


def load(path: ty.Union[pathlib.Path, str]) -> gn.Graph:
    """
    Shorthand for ``yape.gn.Graph.load(path)``.
    """
    return gn.Graph.load(path)


def gr(nodegen: ty.Callable,
       /,
       args=None,
       kwargs=None,
       **kw,
       ) -> ty.Union[ty.Callable, gn.Graph]:
    def graph_creator(*nodegen_args, **nodegen_kwargs) -> gn.Graph:
        g = gn.Graph(**kw)
        with g:
            nodegen(*nodegen_args, **nodegen_kwargs)
        return g

    if args is None and kwargs is None:
        return graph_creator

    if args is None:
        args = tuple()

    if kwargs is None:
        kwargs = {}

    return graph_creator(*args, **kwargs)


def fn(f: ty.Any,
       /,
       args=None,
       kwargs=None,
       **kw,
       ) -> ty.Union[ty.Callable, gn.Node]:
    def node_creator(*call_args, **call_kwargs) -> gn.Node:
        op = nodeop.Call(f, call_args, call_kwargs)
        return gn.Node(op, **kw)

    if args is None and kwargs is None:
        return node_creator

    if args is None:
        args = tuple()

    if kwargs is None:
        kwargs = {}

    return node_creator(*args, **kwargs)


def value(v: ty.Any = nodeop.UNSET, /, **kw) -> gn.Node:
    op = nodeop.Value(v)
    return gn.Node(op, **kw)


def data(payload: ty.Any, /, id: str = None, **kw) -> gn.Node:
    op = nodeop.Data(payload, id)
    return gn.Node(op, **kw)


def run(*k, **kw):
    runner = grun.Runner()
    return runner.run(*k, **kw)
